# -*- coding: utf-8 -*-
from __future__ import annotations

from xtyping import Any, Protocol, Union

__all__ = (
    "JsonInterface",
    "JsonInterfaceProperty",
    "JsonInterfaceProtocol",
    "Dumpable",
)


class Dumpable(Protocol):
    def __dumpable__(self) -> Any:
        ...


class JsonInterface(Protocol):
    def __json_interface__(self) -> Any:
        ...


class JsonInterfaceProperty(Protocol):
    @property
    def __json_interface__(self) -> Any:
        ...


JsonInterfaceProtocol = Union[JsonInterface, JsonInterfaceProperty]



# -*- coding: utf-8 -*-
"""jsonbourne jsonlib api"""
# pyright: reportOptionalMemberAccess=false
from __future__ import annotations

import json as pyjson

from abc import ABC, abstractmethod
from datetime import date as dtdate, datetime, time as dttime, timedelta
from decimal import Decimal
from pathlib import Path
from sys import modules as _sys_modules
from typing import Any, Callable, List, Optional, Tuple, Type, Union
from uuid import UUID

from jsonbourne.protocols import Dumpable, JsonInterfaceProtocol

try:
    import dataclasses
except ImportError:
    dataclasses = None  # type: ignore[assignment]

try:
    import rapidjson

    RapidJSONEncoder = rapidjson.Encoder
    RapidJSONDecoder = rapidjson.Decoder
    RapidJSONDecodeError = rapidjson.JSONDecodeError

except ImportError:
    rapidjson = None
    RapidJSONEncoder = None
    RapidJSONDecoder = None
    RapidJSONDecodeError = None

try:
    import orjson

except ImportError:
    orjson = None  # type: ignore[assignment]

try:
    import numpy as np
except ImportError:
    np = None

JSONLIB_DEFAULT_PREFERENCE = (
    "orjson",
    "rapidjson",
)


def _json_interface(obj: JsonInterfaceProtocol) -> Any:
    _json_interface = obj.__json_interface__
    if callable(_json_interface):
        _json_interface = _json_interface()
    return _json_interface


def _dumpable(obj: Dumpable) -> Any:
    return obj.__dumpable__()


def _json_encode_default(obj: Any) -> Any:
    if hasattr(obj, "__dumpable__"):
        return _dumpable(obj)
    if hasattr(obj, "__json_interface__"):
        return _json_interface(obj)
    if np:
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, (np.ndarray, np.generic)):
            return obj.tolist()
    if dataclasses is not None:
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
    if isinstance(obj, bytes):
        return str(obj, encoding="utf-8")
    if isinstance(obj, tuple):
        return tuple(obj)
    if isinstance(obj, Path):
        return str(obj)
    if isinstance(obj, UUID):
        return str(obj)
    if isinstance(obj, (datetime, dttime, dtdate)):
        return obj.isoformat()
    if isinstance(obj, timedelta):
        return obj.total_seconds()
    if isinstance(obj, Decimal):
        return float(obj)
    if hasattr(obj, "eject"):
        return obj.eject()
    if hasattr(obj, "to_dict"):
        return obj.to_dict()
    if hasattr(obj, "dict"):
        return obj.dict()

    raise TypeError("Cannot encode obj as JSON: {}".format(str(obj)))


class JsonLibABC(ABC):
    lib = "json"

    @staticmethod
    @abstractmethod
    def dumps(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        ...

    @staticmethod
    @abstractmethod
    def dumpb(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        ...

    @staticmethod
    @abstractmethod
    def loads(string: Union[bytes, str], jsonc: bool = False, **kwargs: Any) -> Any:
        ...

    @staticmethod
    @abstractmethod
    def useable() -> bool:
        ...

    @staticmethod
    def default(obj: Any) -> Any:
        """Default encoder"""
        return _json_encode_default(obj)

    @staticmethod
    def has_rapidjson() -> bool:
        return rapidjson is not None

    @staticmethod
    def has_orjson() -> bool:
        return orjson is not None

    @staticmethod
    @abstractmethod
    def jsoncp(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> Any:
        ...


class JSON_STDLIB(JsonLibABC):
    lib = "json"
    JSONEncoder = pyjson.JSONEncoder
    JSONDecoder = pyjson.JSONDecoder
    JSONDecodeError = pyjson.JSONDecodeError

    @staticmethod
    def dumps(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        _pretty = fmt or pretty
        separators = (",", ": ") if _pretty else (",", ":")
        dump_str = pyjson.dumps(
            data,
            indent=2 if _pretty else None,
            sort_keys=sort_keys,
            separators=separators,
            default=default or _json_encode_default,
            **kwargs,
        )
        if append_newline:
            return f"{dump_str}\n"
        return dump_str

    @staticmethod
    def dumpb(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        return JSON_STDLIB.dumps(
            data,
            fmt=fmt,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default or _json_encode_default,
            **kwargs,
        ).encode()

    @staticmethod
    def loads(string: Union[bytes, str], jsonc: bool = False, **kwargs: Any) -> Any:
        if jsonc:
            raise NotImplementedError(
                "Python stdlib-json does not support JSONC; use rapidjson"
            )
        return pyjson.loads(string, **kwargs)

    @staticmethod
    def useable() -> bool:
        return True

    @staticmethod
    @abstractmethod
    def jsoncp(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> Any:
        return JSON_STDLIB.loads(
            JSON_STDLIB.dumps(
                data,
                fmt=fmt,
                pretty=pretty,
                sort_keys=sort_keys,
                append_newline=append_newline,
                default=default or _json_encode_default,
                **kwargs,
            )
        )


class ORJSON(JsonLibABC):
    lib = "orjson"

    @staticmethod
    def dumps(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        return ORJSON.dumpb(
            data,
            fmt=fmt,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default or _json_encode_default,
        ).decode(encoding="utf-8")

    @staticmethod
    def dumpb(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        option = 0
        if fmt or pretty:
            option |= orjson.OPT_INDENT_2
        if sort_keys:
            option |= orjson.OPT_SORT_KEYS
        if append_newline:
            option |= orjson.OPT_APPEND_NEWLINE
        if "numpy" in _sys_modules:
            option |= orjson.OPT_SERIALIZE_NUMPY

        return orjson.dumps(
            data, option=option, default=default or _json_encode_default
        )

    @staticmethod
    def loads(string: Union[bytes, str], jsonc: bool = False, **kwargs: Any) -> Any:
        if jsonc:
            raise NotImplementedError("orjson does not support JSONC; use rapidjson")
        return orjson.loads(string)

    @staticmethod
    def useable() -> bool:
        return JsonLibABC.has_orjson()

    @staticmethod
    @abstractmethod
    def jsoncp(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> Any:
        return ORJSON.loads(
            ORJSON.dumpb(
                data,
                fmt=fmt,
                pretty=pretty,
                sort_keys=sort_keys,
                append_newline=append_newline,
                default=default or _json_encode_default,
                **kwargs,
            )
        )


class RAPIDJSON(JsonLibABC):
    lib = "rapidjson"

    JSONEncoder = RapidJSONEncoder
    JSONDecoder = RapidJSONDecoder
    JSONDecodeError = RapidJSONDecodeError

    @staticmethod
    def dumps(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        dump_str = str(
            rapidjson.dumps(
                data,
                indent=2 if (fmt or pretty) else None,
                sort_keys=sort_keys,
                default=default or _json_encode_default,
                datetime_mode=rapidjson.DM_ISO8601,
                number_mode=rapidjson.NM_NAN,
                uuid_mode=rapidjson.UM_CANONICAL,
                **kwargs,
            )
        )
        if append_newline:
            return f"{dump_str}\n"
        return dump_str

    @staticmethod
    def dumpb(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        return RAPIDJSON.dumps(
            data,
            fmt=fmt,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default or _json_encode_default,
            **kwargs,
        ).encode()

    @staticmethod
    def loads(string: Union[bytes, str], jsonc: bool = False, **kwargs: Any) -> Any:
        if jsonc and "parse_mode" not in kwargs:
            kwargs["parse_mode"] = rapidjson.PM_COMMENTS
        return rapidjson.loads(string, **kwargs)

    @staticmethod
    def useable() -> bool:
        return JsonLibABC.has_rapidjson()

    @staticmethod
    @abstractmethod
    def jsoncp(
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> Any:
        return RAPIDJSON.loads(
            RAPIDJSON.dumps(
                data,
                fmt=fmt,
                pretty=pretty,
                sort_keys=sort_keys,
                append_newline=append_newline,
                default=default or _json_encode_default,
                **kwargs,
            )
        )


def pick_lib() -> "Type[JsonLibABC]":
    if ORJSON.useable():
        return ORJSON
    if RAPIDJSON.useable():
        return RAPIDJSON
    return JSON_STDLIB


def _import_rapidjson() -> "Type[RAPIDJSON]":
    if RAPIDJSON.useable():
        return RAPIDJSON
    raise ImportError("rapidjson (python-rapidjson) not installed")


def _import_orjson() -> "Type[ORJSON]":
    if ORJSON.useable():
        return ORJSON
    raise ImportError("orjson not installed")


def _import_json_stdlib() -> "Type[JSON_STDLIB]":
    return JSON_STDLIB


def import_json(
    jsonlibs: Optional[Union[Tuple[str, ...], List[str]]] = None
) -> "Type[JsonLibABC]":
    lib2funk = {
        "rapidjson": _import_rapidjson,
        "orjson": _import_orjson,
        "json": _import_json_stdlib,
    }
    jsonlibs = jsonlibs or JSONLIB_DEFAULT_PREFERENCE

    for mod in jsonlibs:
        try:
            return lib2funk[mod]()
        except (ImportError, ModuleNotFoundError):
            pass

    return _import_json_stdlib()


class JsonLib:
    _jsonlib: Type[JsonLibABC]
    _oj: Optional[Type[ORJSON]] = None
    _rj: Optional[Type[RAPIDJSON]] = None

    def __init__(self, jsonlib: Optional[Type[JsonLibABC]] = None):
        if jsonlib:
            self._jsonlib = jsonlib
        else:
            self._jsonlib = import_json()
        if ORJSON.useable():
            self._oj = ORJSON
        if RAPIDJSON.useable():
            self._rj = RAPIDJSON

    def dumps(
        self,
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> str:
        return self._jsonlib.dumps(
            data=data,
            fmt=fmt,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default,
            **kwargs,
        )

    def dumpb(
        self,
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> bytes:
        return self._jsonlib.dumpb(
            data=data,
            fmt=fmt,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default,
            **kwargs,
        )

    def loads(
        self, string: Union[bytes, str], jsonc: bool = False, **kwargs: Any
    ) -> Any:
        if jsonc:
            if self._rj:
                return self._rj.loads(string, jsonc=jsonc, **kwargs)
        return self._jsonlib.loads(string, jsonc=jsonc, **kwargs)

    def orjson_useable(self) -> bool:
        return ORJSON.useable()

    def rapidjson_useable(self) -> bool:
        return RAPIDJSON.useable()

    def use_orjson(self) -> None:
        self._jsonlib = _import_orjson()

    def use_rapidjson(self) -> None:
        self._jsonlib = _import_rapidjson()

    def use_json_stdlib(self) -> None:
        self._jsonlib = _import_json_stdlib()

    def use_json(self) -> None:
        self.use_json_stdlib()

    def which(self) -> str:
        return self._jsonlib.lib

    def jsoncp(
        self,
        data: Any,
        fmt: bool = False,
        pretty: bool = False,
        sort_keys: bool = False,
        append_newline: bool = False,
        default: Optional[Callable[[Any], Any]] = None,
        **kwargs: Any,
    ) -> Any:
        return self._jsonlib.jsoncp(
            data=data,
            fmt=fmt,
            pretty=pretty,
            sort_keys=sort_keys,
            append_newline=append_newline,
            default=default,
            **kwargs,
        )


_JSONLIB: Type[JsonLibABC] = import_json()
JSONLIB = JsonLib(_JSONLIB)


def dumps(
    data: Any,
    fmt: bool = False,
    pretty: bool = False,
    sort_keys: bool = False,
    append_newline: bool = False,
    default: Optional[Callable[[Any], Any]] = None,
    **kwargs: Any,
) -> str:
    return JSONLIB.dumps(
        data=data,
        fmt=fmt,
        pretty=pretty,
        sort_keys=sort_keys,
        append_newline=append_newline,
        default=default,
        **kwargs,
    )


def dumpb(
    data: Any,
    fmt: bool = False,
    pretty: bool = False,
    sort_keys: bool = False,
    append_newline: bool = False,
    default: Optional[Callable[[Any], Any]] = None,
    **kwargs: Any,
) -> bytes:
    return JSONLIB.dumpb(
        data=data,
        fmt=fmt,
        pretty=pretty,
        sort_keys=sort_keys,
        append_newline=append_newline,
        default=default,
        **kwargs,
    )


def loads(string: Union[bytes, str], jsonc: bool = False, **kwargs: Any) -> Any:
    return JSONLIB.loads(string, jsonc=jsonc, **kwargs)


def parse(string: Union[bytes, str], jsonc: bool = False, **kwargs: Any) -> Any:
    return JSONLIB.loads(string, jsonc=jsonc, **kwargs)


def jsoncp(
    data: Any,
    fmt: bool = False,
    pretty: bool = False,
    sort_keys: bool = False,
    append_newline: bool = False,
    default: Optional[Callable[[Any], Any]] = None,
    **kwargs: Any,
) -> Any:
    return JSONLIB.jsoncp(
        data=data,
        fmt=fmt,
        pretty=pretty,
        sort_keys=sort_keys,
        append_newline=append_newline,
        default=default,
        **kwargs,
    )


def orjson_useable() -> bool:
    return JSONLIB.orjson_useable()


def rapidjson_useable() -> bool:
    return JSONLIB.rapidjson_useable()


def use_orjson() -> None:
    JSONLIB.use_orjson()


def use_rapidjson() -> None:
    JSONLIB.use_rapidjson()


def use_json_stdlib() -> None:
    JSONLIB.use_json_stdlib()


def use_json() -> None:
    use_json_stdlib()


def which() -> str:
    return JSONLIB.which()



# -*- coding: utf-8 -*-
from __future__ import annotations

import os

from typing import Any

from jsonbourne import JSON

try:
    from starlette.responses import Response
except ModuleNotFoundError:
    if not ("CI" in os.environ and os.environ["CI"] == "true"):
        raise ModuleNotFoundError(
            "starlette not found/installed; `pip install starlette`"
        )
    else:
        Response = object  # type: ignore[assignment, misc]

__all__ = ("JSONBOURNEResponse",)


class JSONBOURNEResponse(Response):
    """FastAPI/starlette json response to auto use jsonbourne"""

    media_type = "application/json"

    def render(self, content: Any) -> bytes:
        """Return JSON string for content as bytes"""
        return JSON.binify(data=content)



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import cast, overload

from xtyping import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    MutableSequence,
    Optional,
    Protocol,
    SupportsIndex,
    Tuple,
    Type,
    TypeVar,
    Union,
)

__all__ = (
    "JsonArr",
    "n_args",
)
_T = TypeVar("_T")
_R = TypeVar("_R")


class SupportsDunderLT(Protocol):
    def __lt__(self, __other: Any) -> bool:
        ...


class SupportsDunderGT(Protocol):
    def __gt__(self, __other: Any) -> bool:
        ...


SupportsRichComparison = Union[SupportsDunderLT, SupportsDunderGT]
SupportsRichComparisonT = TypeVar(
    "SupportsRichComparisonT", bound=SupportsRichComparison
)  # noqa: Y001


def n_args(fn: Callable[..., _R]) -> int:
    if not callable(fn):
        raise TypeError(f"{fn!r} is not callable")
    try:
        return fn.__code__.co_argcount
    except AttributeError:
        pass
    if hasattr(fn, "__call__") and callable(fn):
        _nargs = n_args(fn.__call__)  # type: ignore[operator]
        co_varnames = fn.__call__.__code__.co_varnames  # type: ignore[operator]
        if co_varnames[0] == "self":
            return len(co_varnames) - 1
        return _nargs
    raise TypeError(f"{fn!r} is not callable")


class JsonArr(MutableSequence[_T], Generic[_T]):
    __arr: List[_T]

    def __init__(self, iterable: Optional[Iterable[_T]] = None) -> None:
        self.__arr = list(iterable or [])

    def __post_init__(self) -> Any:
        """Function place holder that is called after object initialization"""
        pass  # pylint: disable=unnecessary-pass

    def __repr__(self) -> str:
        return "JsonArr(" + repr(self.__arr) + ")"

    def __str__(self) -> str:
        return "JsonArr({})".format(self.__arr.__str__())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, JsonArr):
            return self.__arr == other.__arr
        if isinstance(other, list):
            return self.__arr == other
        if isinstance(other, tuple):
            return self.__arr == list(other)
        return False

    def copy(self) -> JsonArr[_T]:
        return JsonArr(self.__arr.copy())

    def append(self, __object: _T) -> None:
        self.__arr.append(__object)

    def extend(self, __iterable: Iterable[_T]) -> None:
        self.__arr.extend(__iterable)

    def push(self, *__object: _T) -> None:
        self.extend(__object)

    def pop(self, index: int = -1) -> _T:
        return self.__arr.pop(index)

    # # Signature of `list.index` should be kept in line with `collections.UserList.index()`
    def index(
        self, __value: _T, __start: SupportsIndex = 0, __stop: SupportsIndex = -1
    ) -> int:
        return self.__arr.index(__value, __start, __stop)

    def count(self, __value: _T) -> int:
        return self.__arr.count(__value)

    def insert(self, __index: SupportsIndex, __object: _T) -> None:
        return self.__arr.insert(__index, __object)

    def remove(self, __value: _T) -> None:
        return self.__arr.remove(__value)

    @overload
    def sort(
        self: JsonArr[SupportsRichComparisonT], *, key: None = None, reverse: bool = ...
    ) -> None:
        ...

    @overload
    def sort(
        self, *, key: Callable[[_T], SupportsRichComparison], reverse: bool = ...
    ) -> None:
        ...

    def sort(
        self,
        *,
        key: Optional[Callable[[_T], SupportsRichComparison]] = None,
        reverse: bool = False,
    ) -> None:
        return self.__arr.sort(key=key, reverse=reverse)

    def __len__(self) -> int:
        return len(self.__arr)

    def __iter__(self) -> Iterator[_T]:
        return iter(self.__arr)

    __hash__: None  # type: ignore[assignment]

    @overload
    def __getitem__(self, __i: SupportsIndex) -> _T:
        ...

    @overload
    def __getitem__(self, __s: slice) -> JsonArr[_T]:
        ...

    def __getitem__(self, __i: Union[SupportsIndex, slice]) -> Union[_T, JsonArr[_T]]:
        if isinstance(__i, slice):
            return JsonArr(self.__arr[__i])
        return self.__arr[__i]

    @overload
    def __setitem__(self, __i: SupportsIndex, __o: _T) -> None:
        ...

    @overload
    def __setitem__(self, __s: slice, __o: Iterable[_T]) -> None:
        ...

    def __setitem__(self, __i: Any, __o: Any) -> None:
        return self.__arr.__setitem__(__i, __o)

    def __delitem__(self, __i: Union[SupportsIndex, slice]) -> None:
        self.__arr.__delitem__(__i)

    def __add__(self, __x: Union[Iterable[_T], JsonArr[_T]]) -> JsonArr[_T]:
        if isinstance(__x, JsonArr):
            return JsonArr(self.__arr + __x.__arr)
        return JsonArr(self.__arr + list(__x))

    def __iadd__(self, __x: Iterable[_T]) -> JsonArr[_T]:
        self.__arr += __x
        return self

    def __mul__(self, __n: int) -> JsonArr[_T]:
        return JsonArr(self.__arr * __n)

    def __rmul__(self, __n: SupportsIndex) -> JsonArr[_T]:
        return JsonArr(self.__arr.__mul__(__n))

    def __imul__(self, __n: SupportsIndex) -> JsonArr[_T]:
        self.__arr.__mul__(__n)
        return self

    def __contains__(self, __o: object) -> bool:
        return self.__arr.__contains__(__o)

    def __reversed__(self) -> Iterator[_T]:
        return reversed(self.__arr)

    def __gt__(self, __x: List[_T]) -> bool:
        return self.__arr > __x

    def __ge__(self, __x: List[_T]) -> bool:
        return self.__arr >= __x

    def __lt__(self, __x: List[_T]) -> bool:
        return self.__arr < __x

    def __le__(self, __x: List[_T]) -> bool:
        return self.__arr <= __x

    def as_list(self) -> List[_T]:
        return list(self.__arr)

    def as_tuple(self) -> Tuple[_T, ...]:
        return tuple(self.__arr)

    @classmethod
    def validate_type(cls: Type[JsonArr[_T]], val: Any) -> JsonArr[_T]:
        """Validate and convert a value to a JsonObj object"""
        return cls(val)

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable[[Any], Any]]:
        """Return the JsonObj validator functions"""
        yield cls.validate_type

    def eject(self) -> List[_T]:
        return self.__arr

    @overload
    def enumerate(
        self, start: int = 0, flip: Literal[False] = False
    ) -> Iterator[Tuple[int, _T]]:
        ...

    @overload
    def enumerate(
        self, start: int = 0, flip: Literal[True] = True
    ) -> Iterator[Tuple[_T, int]]:
        ...

    @overload
    def enumerate(
        self, start: int = 0, flip: bool = False
    ) -> Union[Iterator[Tuple[int, _T]], Iterator[Tuple[_T, int]]]:
        ...

    def enumerate(
        self, start: int = 0, flip: bool = False
    ) -> Union[Iterator[Tuple[int, _T]], Iterator[Tuple[_T, int]]]:
        ...
        if flip:
            return zip(self.__arr, range(start, len(self.__arr) + start))
        return enumerate(self.__arr, start=start)

    def _iter_el(self) -> Iterator[_T]:
        return iter(self.__arr)

    def _iter_el_ix(self) -> Iterator[Tuple[_T, int]]:
        return self.enumerate(flip=True)

    def _iter_el_ix_arr(self) -> Iterator[Tuple[_T, int, JsonArr[_T]]]:
        return ((el, ix, self) for el, ix in self.enumerate(flip=True))

    def _filter_el(self, func: Callable[[_T], bool]) -> JsonArr[_T]:
        return JsonArr(list(filter(func, self._iter_el())))

    def _filter_el_ix(self, func: Callable[[_T, int], bool]) -> JsonArr[_T]:
        return JsonArr([el for el, ix in self._iter_el_ix() if func(el, ix)])

    def _filter_el_ix_arr(
        self, func: Callable[[_T, int, JsonArr[_T]], bool]
    ) -> JsonArr[_T]:
        return JsonArr(
            [el for el, ix, arr in self._iter_el_ix_arr() if func(el, ix, arr)]
        )

    def filter(
        self,
        func: Union[
            Callable[[_T], bool],
            Callable[[_T, int], bool],
            Callable[[_T, int, JsonArr[_T]], bool],
        ],
        nargs: Optional[Union[Literal[1], Literal[2], Literal[3]]] = None,
    ) -> JsonArr[_T]:
        _fn_args = nargs or n_args(func)
        if _fn_args == 3:
            _fn3 = cast(Callable[[_T, int, JsonArr[_T]], bool], func)
            return self._filter_el_ix_arr(_fn3)
        elif _fn_args == 2:
            _fn2 = cast(Callable[[_T, int], bool], func)
            return self._filter_el_ix(_fn2)
        elif _fn_args == 1:
            _fn1 = cast(Callable[[_T], bool], func)
            return self._filter_el(_fn1)
        raise TypeError("Could not determine number of arguments for filter function")

    def _map_el(self, func: Callable[[_T], _R]) -> JsonArr[_R]:
        return JsonArr(list(map(func, self._iter_el())))

    def _map_el_ix(self, func: Callable[[_T, int], _R]) -> JsonArr[_R]:
        return JsonArr(list(map(lambda el_ix: func(*el_ix), self._iter_el_ix())))

    def _map_el_ix_arr(self, func: Callable[[_T, int, JsonArr[_T]], _R]) -> JsonArr[_R]:
        return JsonArr(
            list(map(lambda el_ix: func(el_ix[0], el_ix[1], self), self._iter_el_ix()))
        )

    def map(
        self,
        func: Union[
            Callable[[_T], _R],
            Callable[[_T, int], _R],
            Callable[[_T, int, JsonArr[_T]], _R],
        ],
        nargs: Optional[Union[Literal[1], Literal[2], Literal[3]]] = None,
    ) -> JsonArr[_R]:
        _fn_args = nargs or n_args(func)
        if _fn_args == 3:
            _fn3 = cast(Callable[[_T, int, JsonArr[_T]], _R], func)
            return self._map_el_ix_arr(_fn3)
        elif _fn_args == 2:
            _fn2 = cast(Callable[[_T, int], _R], func)
            return self._map_el_ix(_fn2)
        elif _fn_args == 1:
            _fn1 = cast(Callable[[_T], _R], func)
            return self._map_el(_fn1)
        raise TypeError("Could not determine number of arguments for map function")

    def slice(
        self, start: Optional[int] = None, end: Optional[int] = None
    ) -> JsonArr[_T]:
        if start is None and end is None:
            return self
        _start = 0 if start is None else start
        _end = len(self) if end is None else end
        return JsonArr(self.__arr[_start:_end])



# -*- coding: utf-8 -*-
from __future__ import annotations

from jsonbourne.protocols import (
    Dumpable,
    JsonInterface,
    JsonInterfaceProperty,
    JsonInterfaceProtocol,
)

__all__ = (
    "JsonInterface",
    "JsonInterfaceProperty",
    "JsonInterfaceProtocol",
    "Dumpable",
)



# -*- coding: utf-8 -*-
"""JSON callable module ~ jsonbourne"""
from __future__ import annotations

import sys

from jsonbourne import __version__, json
from jsonbourne.core import (
    JSON as _JSON,
    UNDEFINED as UNDEFINED,
    JsonDict as JsonDict,
    JSONModuleCls,
    JsonObj as JsonObj,
    JsonObjMutableMapping,
    Null as Null,
    null as null,
    undefined as undefined,
)
from jsonbourne.jsonlib import import_json

stringify = _JSON.stringify
dumps = _JSON.dumps
binify = _JSON.binify
dumpb = _JSON.dumpb
loads = _JSON.loads
parse = _JSON.parse
jsoncp = _JSON.jsoncp
json_lib = _JSON.json_lib
which = _JSON.which
use_orjson = _JSON.use_orjson
use_rapidjson = _JSON.use_rapidjson
use_json_stdlib = _JSON.use_json_stdlib
jsonify = _JSON.jsonify
unjsonify = _JSON.unjsonify
JSONDecodeError = _JSON.JSONDecodeError

__all__ = (
    "JSONModuleCls",
    "JsonDict",
    "JsonObj",
    "JsonObjMutableMapping",
    "Null",
    "UNDEFINED",
    "__version__",
    "binify",
    "dumpb",
    "dumps",
    "import_json",
    "json",
    "json_lib",
    "jsoncp",
    "jsonify",
    "loads",
    "null",
    "parse",
    "stringify",
    "undefined",
    "unjsonify",
    "use_json_stdlib",
    "use_orjson",
    "use_rapidjson",
)

# if __name__ is NOT __main__, funkify module
if __name__ != "__main__":
    sys.modules["JSON"].__class__ = JSONModuleCls



# -*- coding: utf-8 -*-
"""trydantic = try + pydantic"""
from __future__ import annotations

from dataclasses import (
    MISSING,
    Field,
    FrozenInstanceError,
    InitVar,
    asdict,
    astuple,
    field,
    fields,
    is_dataclass,
    make_dataclass,
    replace,
)

PYDANTIC_AVAILABLE: bool = False

try:
    # use `pydantic.dataclasses.dataclass` if available
    from pydantic.dataclasses import dataclass as dataclass

    PYDANTIC_AVAILABLE = True
except ImportError:
    from dataclasses import dataclass as _dataclass
    from functools import wraps

    dataclass = wraps(_dataclass)(
        lambda *args, **kwargs: _dataclass(
            *args, **{k: v for k, v in kwargs.items() if k != "config"}
        )
    )

__all__ = (
    "Field",
    "FrozenInstanceError",
    "InitVar",
    "MISSING",
    "PYDANTIC_AVAILABLE",
    "asdict",
    "astuple",
    "dataclass",
    "field",
    "fields",
    "is_dataclass",
    "make_dataclass",
    "replace",
)



# -*- coding: utf-8 -*-
"""JSONBourne helper funks and utils"""
from __future__ import annotations

import re

from typing import Match

__all__ = ("rm_js_comments", "strip_comments", "STRIP_JSON_COMMENTS_REGEX")

STRIP_JSON_COMMENTS_REGEX = r"(\".*?(?<!\\)\"|\'.*?(?<!\\)\')|(/\*.*?\*/|//[^\r\n]*$)"


def _strip_json_comments_regex_replacer(match: Match[str]) -> str:
    # if the 2nd group (capturing comments) is not None,
    # it means we have captured a non-quoted (real) comment string.
    if match.group(2) is not None:
        return ""  # so we will return empty to remove the comment
    return match.group(1)  # captured quoted-string


def rm_js_comments(string: str) -> str:
    """Rejects/regex that removes js/ts/json style comments

    Source (stackoverflow):
        https://stackoverflow.com/a/18381470
    """
    # first group captures quoted strings (double or single)
    # second group captures comments (//single-line or /* multi-line */)
    regex = re.compile(STRIP_JSON_COMMENTS_REGEX, re.MULTILINE | re.DOTALL)

    return regex.sub(_strip_json_comments_regex_replacer, string)


def strip_comments(string: str) -> str:

    return rm_js_comments(string)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['jsonbourne', 'jsonbourne.dev']

package_data = \
{'': ['*']}

modules = \
['JSON', 'david_webb']
install_requires = \
['xtyping>=0.5.0']

extras_require = \
{'all': ['pydantic>=1.8.2', 'python-rapidjson>=0.9.1', 'orjson>=3.0.0,<4.0.0'],
 'full': ['pydantic>=1.8.2', 'python-rapidjson>=0.9.1', 'orjson>=3.0.0,<4.0.0'],
 'orjson': ['orjson>=3.0.0,<4.0.0'],
 'pydantic': ['pydantic>=1.8.2'],
 'rapidjson': ['python-rapidjson>=0.9.1']}

setup_kwargs = {
    'name': 'jsonbourne',
    'version': '0.22.2',
    'description': 'EZPZ JSON',
    'long_description': '<a href="https://github.com/dynamic-graphics-inc/dgpy-libs">\n<img align="right" src="https://github.com/dynamic-graphics-inc/dgpy-libs/blob/main/docs/images/dgpy_banner.svg?raw=true" alt="drawing" height="120" width="300"/>\n</a>\n\n# jsonbourne\n\n[![Wheel](https://img.shields.io/pypi/wheel/jsonbourne.svg)](https://img.shields.io/pypi/wheel/jsonbourne.svg)\n[![Version](https://img.shields.io/pypi/v/jsonbourne.svg)](https://img.shields.io/pypi/v/jsonbourne.svg)\n[![py_versions](https://img.shields.io/pypi/pyversions/jsonbourne.svg)](https://img.shields.io/pypi/pyversions/jsonbourne.svg)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**Install:** `pip install jsonbourne`\n\n- Python json lib/pkg that makes json feel like the JSON module in javascript/typescript:\n  - `from jsonbourne import JSON; JSON.parse(JSON.stringify({"key": "value"}))`\n  - Automatically uses best json-lib-backend avalible (`orjson`/`python-rapidjson`) ~ can be configured\n- Hybrid dict/class object (`jsonbourne.JsonObj`):\n  - Dot-notation getting/setting (featuring protected attributes!)\n  - All your favorite python dictionary methods (`items`, `keys`, `update`, `values`) and more!\n  - Works with `pydantic` and `attrs`\n- FastAPI:\n  - JSONBOURNEResponse ~ auto use the best\n- No hard dependencies ~ works with python-stdlib-json as well as `orjson` and `python-rapidjson`\n- `jsonbourne.JsonObj` uses list/dict comprehensions (some are recursive) everywhere because \'why not?\' and it is a bit faster\n\n## Usage:\n\n### JSON ~ `from jsonbourne import JSON`\n\n**Importing:**\n\n```python\n# Importing JSON:\nfrom jsonbourne import JSON\n\n# or\nimport JSON\n\n# Importing jsonbourne:\nimport jsonbourne\nimport david_webb  # jsonbourne\'s `True` identity\n```\n\n**JSON basics:**\n\n```python\nimport JSON  # Module included with jsonbourne\n\nstring_stringify = JSON.stringify(\n    {"a": 1, "b": 2, "c": 3}\n)  # \'{"a": 1, "b": 2, "c": 3}\'\nstring_dumps = JSON.dumps({"a": 1, "b": 2, "c": 3})  # \'{"a": 1, "b": 2, "c": 3}\'\nstring_dumps\n```\n\n    \'{"a":1,"b":2,"c":3}\'\n\n#### JSON option kwargs ~ `pretty` & `sort_keys`\n\n**pretty:**\n\n```python\nstring_dumps = JSON.stringify(\n    {"b": 2, "a": 1, "c": 3}, pretty=True\n)  # \'{"a": 1, "b": 2, "c": 3}\'\nprint(string_dumps)\n```\n\n    {\n      "b": 2,\n      "a": 1,\n      "c": 3\n    }\n\n**sort_keys:**\n\n```python\nstring_dumps = JSON.stringify(\n    {"b": 2, "a": 1, "c": 3}, pretty=True, sort_keys=True\n)  # \'{"a": 1, "b": 2, "c": 3}\'\nprint(string_dumps)\n```\n\n    {\n      "a": 1,\n      "b": 2,\n      "c": 3\n    }\n\n### JsonObj & JSON\n\n- Python dictionary/object with dot access\n- Protections against setting class/obj attributes\n- Is as javascript-y as possible (keys have to be strings -- ints/floats will be converted to strings)\n- Create a `jsonbourne.JsonObj` with `jsonbourne.JSON`\n- Recursive jsonification\n- Allows for kwarging (`**json_obj`)\n- Works with `pydantic` and `attrs`\n\n#### Make an empty JsonObj\n\nThe following 3 examples all produce the same thing\n\n```python\nfrom jsonbourne import JSON\nj = JSON()  # j => JsonObj(**{})\n# OR\nimport JSON\nj = JSON()  # j => JsonObj(**{})\n# OR\nfrom jsonbourne import JsonObj\nj = JsonObj()  # j => JsonObj(**{})\n```\n\n#### From a dictionary o data\n\n```python\nimport datetime\n\ndata = {\n    "key": "value",\n    "list": [1, 2, 3, 4, 5],\n    "dt": datetime.datetime(1970, 1, 1, 0, 0, 0, 1),\n    "sub": {\n        "b": 3,\n        "key": "val",\n        "a": 1,\n    },\n    "timedelta": datetime.timedelta(days=2),\n}\n\nJSON(data)\n```\n\n<pre>JsonObj(**{\n    \'dt\': datetime.datetime(1970, 1, 1, 0, 0, 0, 1),\n    \'key\': \'value\',\n    \'list\': [1, 2, 3, 4, 5],\n    \'sub\': {\'a\': 1, \'b\': 3, \'key\': \'val\'},\n    \'timedelta\': datetime.timedelta(days=2)\n})</pre>\n\n### Dot access\n\n```python\nJSON(data).sub.b\n```\n\n    3\n\n```python\nstringified_data = JSON(data).stringify(pretty=True)\nprint(stringified_data)\n```\n\n    {\n      "key": "value",\n      "list": [\n        1,\n        2,\n        3,\n        4,\n        5\n      ],\n      "dt": "1970-01-01T00:00:00.000001",\n      "sub": {\n        "b": 3,\n        "key": "val",\n        "a": 1\n      },\n      "timedelta": 172800.0\n    }\n\n```python\nparsed_data = JSON(stringified_data)\nparsed_data\n```\n\n<pre>JsonObj(**{\n    \'dt\': \'1970-01-01T00:00:00.000001\',\n    \'key\': \'value\',\n    \'list\': [1, 2, 3, 4, 5],\n    \'sub\': {\'a\': 1, \'b\': 3, \'key\': \'val\'},\n    \'timedelta\': 172800.0\n})</pre>\n\n```python\nlist(parsed_data.keys())\n```\n\n    [\'key\', \'list\', \'dt\', \'sub\', \'timedelta\']\n\n```python\nlist(parsed_data.items())\n```\n\n    [(\'key\', \'value\'),\n     (\'list\', [1, 2, 3, 4, 5]),\n     (\'dt\', \'1970-01-01T00:00:00.000001\'),\n     (\'sub\', JsonObj(**{\'b\': 3, \'key\': \'val\', \'a\': 1})),\n     (\'timedelta\', 172800.0)]\n\n```python\nlist(parsed_data.dot_keys())\n```\n\n    [(\'key\',),\n     (\'list\',),\n     (\'dt\',),\n     (\'sub\', \'b\'),\n     (\'sub\', \'key\'),\n     (\'sub\', \'a\'),\n     (\'timedelta\',)]\n\n```python\nlist(parsed_data.dot_items())\n```\n\n    [((\'key\',), \'value\'),\n     ((\'list\',), [1, 2, 3, 4, 5]),\n     ((\'dt\',), \'1970-01-01T00:00:00.000001\'),\n     ((\'sub\', \'b\'), 3),\n     ((\'sub\', \'key\'), \'val\'),\n     ((\'sub\', \'a\'), 1),\n     ((\'timedelta\',), 172800.0)]\n\n```python\nparsed_data[("sub", "key")]\n```\n\n    \'val\'\n\n```python\nparsed_data.dot_lookup("sub.key")\n```\n\n    \'val\'\n\n```python\n{**parsed_data}\n```\n\n    {\'key\': \'value\',\n     \'list\': [1, 2, 3, 4, 5],\n     \'dt\': \'1970-01-01T00:00:00.000001\',\n     \'sub\': JsonObj(**{\'b\': 3, \'key\': \'val\', \'a\': 1}),\n     \'timedelta\': 172800.0}\n\n```python\n# fully eject\nparsed_data.eject()\n```\n\n    {\'key\': \'value\',\n     \'list\': [1, 2, 3, 4, 5],\n     \'dt\': \'1970-01-01T00:00:00.000001\',\n     \'sub\': {\'b\': 3, \'key\': \'val\', \'a\': 1},\n     \'timedelta\': 172800.0}\n\n#### Protected keys\n\n`jsonbourne.JsonObj` protects against setting attributes like `\'items\'` through dot-notation.\n\n```python\nfrom jsonbourne import JSON\n\nj = JSON()\nj.key = "value"\ntry:  # CANNOT set \'items\' using dot-access\n    j.items = [1, 2, 3, 4]\nexcept ValueError:\n    pass\n# CAN set \'items\' through key/item access\nj["items"] = [1, 2, 3, 4]\nprint(j.__dict__)\nprint(j)\nj_items = j.items\nprint("items", j_items)\n# Getting \'items\' through dot-access returns the `items()` method\nassert j.items != [1, 2, 3, 4]\n# Getting \'items\' with key-access returns the stored value\nassert j["items"] == [1, 2, 3, 4]\n```\n\n    {\'_data\': {\'key\': \'value\', \'items\': [1, 2, 3, 4]}}\n    JsonObj(**{\n        \'items\': [1, 2, 3, 4], \'key\': \'value\'\n    })\n    items <bound method JsonObj.items of JsonObj(**{\'key\': \'value\', \'items\': [1, 2, 3, 4]})>\n\n### pydantic & jsonbourne\n\n- `from jsonbourne.pydantic import JsonBaseModel`\n- Allows for aliases when getting/setting attribute(s)\n- Supports `__post_init__` (like dataclasses)\n\n#### Basic usage:\n\n```python\nfrom jsonbourne import JsonObj\nfrom jsonbourne.pydantic import JsonBaseModel\n\n\nclass JsonSubObj(JsonBaseModel):\n    herm: int\n\n    def to_dict(self):\n        return self.dict()\n\n    def to_json(self, *args, **kwargs):\n        return self.json()\n\n    @classmethod\n    def from_json(cls, json_string: str):\n        return JsonSubObj(json.loads(json_string))\n\n\nclass JsonObjModel(JsonBaseModel):\n    a: int\n    b: int\n    c: str\n    d: JsonObj\n    e: JsonSubObj\n\n    #\n    @property\n    def a_property(self) -> str:\n        return "prop_value"\n\n    def to_json(self, *args, **kwargs):\n        return self.json()\n\n    @classmethod\n    def from_json(cls, json_string: str):\n        return cls(**json.loads(json_string))\n\n\nobj = JsonObjModel(\n    **{"a": 1, "b": 2, "c": "herm", "d": {"nested": "nestedval"}, "e": {"herm": 2}}\n)\nobj\n```\n\n<pre>JsonObjModel(**{\n     \'a\': 1,\n     \'b\': 2,\n     \'c\': \'herm\',\n     \'d\': JsonObj(**{\'nested\': \'nestedval\'}),\n     \'e\': {\'herm\': 2}\n})</pre>\n\n```python\n# respects properties (which I don\'t think pydantic does(currently))\nobj.a_property\n```\n\n    \'prop_value\'\n\n---\n\n## JSON backend/lib\n\n**`jsonbourne` finds the best json-lib (python-rapidjson/orjson) it can!** On import `jsonbourne` gets to\nwork spying on your python env. `jsonbourne`, the most highly qualified json-CIA-agent, will import the best\npython-json library it can find; if `jsonbourne`\'s cover is blown (meaning: the\nonly json library found is the python stdlib json), `jsonbourne` will fallback\nto\nthe python stdlib json.\n\n`jsonbourne` will look for the following json-packages in the following order:\n\n1. `rapidjson`\n2. `orjson`\n\n### Custom lib preferences\n\n```python\nfrom jsonbourne import import_json\n\njson = import_json(("rapidjson", "orjson"))  # prefer rapidjson over orjson\nstring = json.dumps({"a": 1, "b": 2, "c": 3})\nprint(json)\nprint(string)\n```\n\n    <class \'jsonbourne.jsonlib.RAPIDJSON\'>\n    {"a":1,"b":2,"c":3}\n\n### Installing better JSON lib:\n\n#### `orjson`\n\n- `pip install orjson` [pip]\n\n#### `rapidjson`/`python-rapidjson`\n\n- `pip install python-rapidjson` [pip]\n- `conda install -c anaconda python-rapidjson` [conda anaconda/defaults]\n- `conda install -c conda-forge python-rapidjson` [conda-forge]\n',
    'author': 'jesse',
    'author_email': 'jesse@dgi.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dynamic-graphics-inc/dgpy-libs/tree/main/libs/jsonbourne',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""Jsonbourne wrapper around httpx clients -- lets you do response.JSON()"""
from __future__ import annotations

from typing import Any

from httpx import AsyncClient, Client, Cookies, Response

from jsonbourne import JSON

__all__ = (
    "Response",
    "AsyncClient",
    "Client",
    "Cookies",
)


def _JSON(self: Response, **kwargs: Any) -> Any:
    return JSON(self.json())


Response.JSON = _JSON



# -*- coding: utf-8 -*-
"""jsonbourne the best undercover json lib

Dynamic Graphics Python
"""
from jsonbourne import jsonlib
from jsonbourne._meta import __version__
from jsonbourne.core import (
    JSON,
    UNDEFINED,
    JsonDict,
    JsonObj,
    JsonObjMutableMapping,
    parse,
    stringify,
    undefined,
)
from jsonbourne.helpers import rm_js_comments
from jsonbourne.jsonlib import import_json

json = jsonlib

JSONLIB = jsonlib.__name__

__all__ = (
    "__version__",
    "JSON",  # js/ts JSON (THE ONE TO USE)
    "jsonlib",  # json compat lib
    "json",  # json compat lib
    # core
    "JsonObjMutableMapping",
    "JsonObj",
    "JsonDict",
    # import
    "import_json",
    # util funks
    "stringify",
    "parse",
    "rm_js_comments",
    # undefined
    "undefined",
    "UNDEFINED",
)
